/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <gdk/gdkinternals.h>
#include "gdkkeysyms.h"
#include "gdkprivate-fb.h"
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>
#include <sys/kd.h>
#include <sys/vt.h>
#include <ncurses.h>
#include <fcntl.h>
#include <unistd.h>


typedef struct _GdkFBKeyboard GdkFBKeyboard;
typedef struct _GdkFBKeyboardDevice GdkFBKeyboardDevice;

struct _GdkFBKeyboard {
  gint fd;
  GIOChannel *io;
  gint io_tag;

  guint modifier_state;
  guint caps_lock : 1;

  gint group;
  gint level;
  
  GdkFBKeyboardDevice *dev;
};

struct _GdkFBKeyboardDevice {
  char *name;
  gboolean (*open)(GdkFBKeyboard *kb);
  void (*close)(GdkFBKeyboard *kb);

  guint    (*lookup_key)               (GdkFBKeyboard       *kb,
					const GdkKeymapKey  *key);
  gboolean (*translate_keyboard_state) (GdkFBKeyboard       *kb,
					guint                hardware_keycode,
					GdkModifierType      state,
					gint                 group,
					guint               *keyval,
					gint                *effective_group,
					gint                *level,
					GdkModifierType     *consumed_modifiers);
  gboolean (*get_entries_for_keyval)   (GdkFBKeyboard       *kb,
					guint                keyval,
					GdkKeymapKey       **keys,
					gint                *n_keys);
  gboolean (*get_entries_for_keycode)  (GdkFBKeyboard       *kb,
					guint                hardware_keycode,
					GdkKeymapKey       **keys,
					guint              **keyvals,
					gint                *n_entries);

  gpointer driver_data;
};

static GdkFBKeyboard *gdk_fb_keyboard = NULL;
static GdkKeymap *default_keymap = NULL;

static gboolean xlate_open            (GdkFBKeyboard       *kb);
static void     xlate_close           (GdkFBKeyboard       *kb);
static guint    xlate_lookup          (GdkFBKeyboard       *kb,
				       const GdkKeymapKey  *key);
static gboolean xlate_translate       (GdkFBKeyboard       *kb,
				       guint                hardware_keycode,
				       GdkModifierType      state,
				       gint                 group,
				       guint               *keyval,
				       gint                *effective_group,
				       gint                *level,
				       GdkModifierType     *consumed_modifiers);
static gboolean xlate_get_for_keyval  (GdkFBKeyboard       *kb,
				       guint                keyval,
				       GdkKeymapKey       **keys,
				       gint                *n_keys);
static gboolean xlate_get_for_keycode (GdkFBKeyboard       *kb,
				       guint                hardware_keycode,
				       GdkKeymapKey       **keys,
				       guint              **keyvals,
				       gint                *n_entries);



static GdkFBKeyboardDevice keyb_devs[] =
{
  {
    "xlate",
    xlate_open,
    xlate_close,
    xlate_lookup,
    xlate_translate,
    xlate_get_for_keyval,
    xlate_get_for_keycode
  },
};

GdkKeymap*
gdk_keymap_get_default_for_display (GdkDisplay *display)
{
  if (default_keymap == NULL)
    default_keymap = g_object_new (gdk_keymap_get_type (), NULL);

  return default_keymap;
}

GdkKeymap*
gdk_keymap_get_for_display (GdkDisplay *display)
{
  return gdk_keymap_get_default_for_display (display);
}

PangoDirection
gdk_keymap_get_direction (GdkKeymap *keymap)
{
  /* FIXME: Only supports LTR keymaps at the moment */
  return PANGO_DIRECTION_LTR;
}

guint
gdk_fb_keyboard_modifiers ()
{
  return gdk_fb_keyboard->modifier_state;
}

gboolean
gdk_fb_keyboard_init (gboolean open_dev)
{
  GdkFBKeyboard *keyb;
  char *keyb_type;
  int i;

  gdk_fb_keyboard = g_new0 (GdkFBKeyboard, 1);
  keyb = gdk_fb_keyboard;
  keyb->fd = -1;
  
  keyb_type = getenv ("GDK_KEYBOARD_TYPE");
  
  if (!keyb_type)
    keyb_type = "xlate";

  for (i = 0; i < G_N_ELEMENTS(keyb_devs); i++)
    {
      if (g_ascii_strcasecmp(keyb_type, keyb_devs[i].name)==0)
	break;
    }
  
  if (i == G_N_ELEMENTS(keyb_devs))
    {
      g_warning ("No keyboard driver of type %s found", keyb_type);
      return FALSE;
    }

  keyb->dev = &keyb_devs[i];

  if (open_dev)
    return gdk_fb_keyboard_open ();
  else
    return TRUE;
}

gboolean
gdk_fb_keyboard_open (void)
{
  GdkFBKeyboard *keyb;
  GdkFBKeyboardDevice *device;

  keyb = gdk_fb_keyboard;
  device = keyb->dev;

  if (!device->open(keyb))
    {
      g_warning ("Keyboard driver open failed");
      return FALSE;
    }

  return TRUE;
}

void 
gdk_fb_keyboard_close (void)
{
  gdk_fb_keyboard->dev->close(gdk_fb_keyboard);
}


gboolean
gdk_keymap_get_entries_for_keyval (GdkKeymap     *keymap,
                                   guint          keyval,
                                   GdkKeymapKey **keys,
                                   gint          *n_keys)
{
  g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), FALSE);
  g_return_val_if_fail (keys != NULL, FALSE);
  g_return_val_if_fail (n_keys != NULL, FALSE);
  g_return_val_if_fail (keyval != 0, FALSE);

  return gdk_fb_keyboard->dev->get_entries_for_keyval (gdk_fb_keyboard,
						       keyval,
						       keys,
						       n_keys);
}

gboolean
gdk_keymap_get_entries_for_keycode (GdkKeymap     *keymap,
                                    guint          hardware_keycode,
                                    GdkKeymapKey **keys,
                                    guint        **keyvals,
                                    gint          *n_entries)
{
  g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), FALSE);
  g_return_val_if_fail (n_entries != NULL, FALSE);
  
  return gdk_fb_keyboard->dev->get_entries_for_keycode (gdk_fb_keyboard,
							hardware_keycode,
							keys,
							keyvals,
							n_entries);
}


guint
gdk_keymap_lookup_key (GdkKeymap          *keymap,
                       const GdkKeymapKey *key)
{
  g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), 0);
  g_return_val_if_fail (key != NULL, 0);
  g_return_val_if_fail (key->group < 4, 0);

  return gdk_fb_keyboard->dev->lookup_key (gdk_fb_keyboard,  key);
}


gboolean
gdk_keymap_translate_keyboard_state (GdkKeymap       *keymap,
                                     guint            hardware_keycode,
                                     GdkModifierType  state,
                                     gint             group,
                                     guint           *keyval,
                                     gint            *effective_group,
                                     gint            *level,
                                     GdkModifierType *consumed_modifiers)
{
  g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), FALSE);
  g_return_val_if_fail (group < 4, FALSE);

  return gdk_fb_keyboard->dev->translate_keyboard_state (gdk_fb_keyboard,
							 hardware_keycode,
							 state,
							 group,
							 keyval,
							 effective_group,
							 level,
							 consumed_modifiers);
}

static void
gdk_fb_handle_key (guint hw_keycode,
		   guint keyval,
		   guint modifier_state,
		   guint8 group,
		   gchar *string,
		   gint string_length,
		   gboolean key_up)
{
  GdkWindow *win;
  GdkEvent *event;

  /* handle some magic keys */
  if (key_up &&
      (modifier_state & GDK_CONTROL_MASK) &&
      (modifier_state & GDK_MOD1_MASK))
    {
      /* FIXME - mozna to ncruses nezvladnout chytit Ctrl Alt Return */	  
      if (keyval == GDK_Return)
	gdk_fb_redraw_all ();
    }

  win = gdk_fb_keyboard_event_window (gdk_fb_window_find_focus (),
				      key_up ? GDK_KEY_RELEASE : GDK_KEY_PRESS);
  if (win)
    {
      event = gdk_event_make (win,
			      key_up ? GDK_KEY_RELEASE : GDK_KEY_PRESS,
			      TRUE);
    
      event->key.state = modifier_state;
      event->key.keyval = keyval;
      event->key.string = string;
      event->key.length = string_length;
      event->key.hardware_keycode = hw_keycode;
      event->key.group = group;
    }
  else
    g_free (string);
}

/******************************************************
 ********* Device specific keyboard code **************
 ******************************************************/


/* XLATE keyboard driver */

struct {
  guint curses_code;  //FIXME
  guint code;
  guint modifier;
} xlate_codes[] =
{
  {KEY_F(1), GDK_F1},
  {KEY_F(2), GDK_F2},
  {KEY_F(3), GDK_F3},
  {KEY_F(4), GDK_F4},
  {KEY_F(5), GDK_F5},
  {KEY_F(6), GDK_F6},
  {KEY_F(7), GDK_F7},
  {KEY_F(8), GDK_F8},
  {KEY_F(9), GDK_F9},
  {KEY_F(10), GDK_F10},

  {KEY_F(11), GDK_F1, GDK_SHIFT_MASK},
  {KEY_F(12), GDK_F2, GDK_SHIFT_MASK},
  {KEY_F(13), GDK_F3, GDK_SHIFT_MASK},
  {KEY_F(14), GDK_F4, GDK_SHIFT_MASK},
  {KEY_F(15), GDK_F5, GDK_SHIFT_MASK},
  {KEY_F(16), GDK_F6, GDK_SHIFT_MASK},
  {KEY_F(17), GDK_F7, GDK_SHIFT_MASK},
  {KEY_F(18), GDK_F8, GDK_SHIFT_MASK},
  {KEY_F(19), GDK_F9, GDK_SHIFT_MASK},
  {KEY_F(20), GDK_F10, GDK_SHIFT_MASK},

  {KEY_HOME, GDK_Home},
  {KEY_PPAGE, GDK_Page_Up}, 
  {KEY_END, GDK_End},
  {KEY_NPAGE, GDK_Page_Down},
  {KEY_IC, GDK_Insert},
  {KEY_DC, GDK_Delete},
  {KEY_BACKSPACE, GDK_BackSpace},

  {KEY_UP, GDK_Up},
  {KEY_LEFT, GDK_Left},
  {KEY_DOWN, GDK_Down},
  {KEY_RIGHT, GDK_Right},
  {KEY_BREAK, GDK_Break},

  {9, GDK_Tab},
  {12, 'l', GDK_CONTROL_MASK},
};

struct {
  guint code;
  guint modifier;
} xlate_chars[] =
{
  /* 0x00 */
  {'@', GDK_CONTROL_MASK},
  {'a', GDK_CONTROL_MASK},
  {'b', GDK_CONTROL_MASK},
  {'c', GDK_CONTROL_MASK},
  {'d', GDK_CONTROL_MASK},
  {'e', GDK_CONTROL_MASK},
  {'f', GDK_CONTROL_MASK},
  {'g', GDK_CONTROL_MASK},
  {'h', GDK_CONTROL_MASK},
  {GDK_Tab, 0},
  {GDK_Return, 0},
  {'k', GDK_CONTROL_MASK},
  {'l', GDK_CONTROL_MASK},
  {'m', GDK_CONTROL_MASK},
  {'n', GDK_CONTROL_MASK},
  {'o', GDK_CONTROL_MASK},
  /* 0x10 */
  {'p', GDK_CONTROL_MASK},
  {'q', GDK_CONTROL_MASK},
  {'r', GDK_CONTROL_MASK},
  {'s', GDK_CONTROL_MASK},
  {'t', GDK_CONTROL_MASK},
  {'u', GDK_CONTROL_MASK},
  {'v', GDK_CONTROL_MASK},
  {'w', GDK_CONTROL_MASK},
  {'x', GDK_CONTROL_MASK},
  {'y', GDK_CONTROL_MASK},
  {'z', GDK_CONTROL_MASK},
  {GDK_Escape, 0},
  {'\\', GDK_CONTROL_MASK},
  {']', GDK_CONTROL_MASK},
  {'^', GDK_CONTROL_MASK},
  {'_', GDK_CONTROL_MASK},
  /* 0x20 */
  {GDK_space, 0},
  {'!', 0},
  {'"', 0},
  {'#', 0},
  {'$', 0},
  {'%', 0},
  {'&', 0},
  {'\'', 0},
  {'(', 0},
  {')', 0},
  {'*', 0},
  {'+', 0},
  {',', 0},
  {'-', 0},
  {'.', 0},
  {'/', 0},
  /* 0x30 */
  {'0', 0},
  {'1', 0},
  {'2', 0},
  {'3', 0},
  {'4', 0},
  {'5', 0},
  {'6', 0},
  {'7', 0},
  {'8', 0},
  {'9', 0},
  {':', 0},
  {';', 0},
  {'<', 0},
  {'=', 0},
  {'>', 0},
  {'?', 0},
  /* 0x40 */
  {'@', 0},
  {'A', 0},
  {'B', 0},
  {'C', 0},
  {'D', 0},
  {'E', 0},
  {'F', 0},
  {'G', 0},
  {'H', 0},
  {'I', 0},
  {'J', 0},
  {'K', 0},
  {'L', 0},
  {'M', 0},
  {'N', 0},
  {'O', 0},
  /* 0x50 */
  {'P', 0},
  {'Q', 0},
  {'R', 0},
  {'S', 0},
  {'T', 0},
  {'U', 0},
  {'V', 0},
  {'W', 0},
  {'X', 0},
  {'Y', 0},
  {'Z', 0},
  {'[', 0},
  {'\\', 0},
  {']', 0},
  {'^', 0},
  {'_', 0},
  /* 0x60 */
  {'`', 0},
  {'a', 0},
  {'b', 0},
  {'c', 0},
  {'d', 0},
  {'e', 0},
  {'f', 0},
  {'g', 0},
  {'h', 0},
  {'i', 0},
  {'j', 0},
  {'k', 0},
  {'l', 0},
  {'m', 0},
  {'n', 0},
  {'o', 0},
  /* 0x70 */
  {'p', 0},
  {'q', 0},
  {'r', 0},
  {'s', 0},
  {'t', 0},
  {'u', 0},
  {'v', 0},
  {'w', 0},
  {'x', 0},
  {'y', 0},
  {'z', 0},
  {'{', 0},
  {'|', 0},
  {'}', 0},
  {'~', 0},
  {GDK_BackSpace, 0},
  /* 0x80 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0x90 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0xA0 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0xB0 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0xC0 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0xD0 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0xE0 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  /* 0xF0 */
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
  {0, 0},
};

void gtk_hierarchy_dump(void);

static gboolean
xlate_io (GIOChannel *gioc,
	  GIOCondition cond,
	  gpointer data)
{
  guint keycode;
  guint modifier;
  gboolean handled;
  int j;

/* Johanka's */

  int key_pressed; 

  while (1) {
    timeout (0);
    key_pressed = getch();

    handled = FALSE;
    modifier = 0;

    if (key_pressed <= 0)
      break;

    if (key_pressed == KEY_MOUSE)
      {
	gdk_fb_mouse_KEY_MOUSE_got();
	continue;
      }

    if (key_pressed == 27)	/* escape? */
      {
	modifier |= GDK_MOD1_MASK;
	timeout (100);	/* msec */
	key_pressed = getch();
	timeout (0);
      }

    for (j = 0; j < G_N_ELEMENTS(xlate_codes); j++)
      {
        if (key_pressed == xlate_codes[j].curses_code)
          {
	    /* Ctrl-Alt Return can't be pressed in the XLATE driver,
	     * use Shift F1 instead */
	    if ((xlate_codes[j].code == GDK_F1) &&
		(xlate_codes[j].modifier & GDK_SHIFT_MASK))
	      {
		gdk_fb_redraw_all ();
		break;
	      }

	    if ((xlate_codes[j].code == 'l') &&
		(xlate_codes[j].modifier & GDK_CONTROL_MASK))
	      {
		gdk_fb_redraw_all ();
		break;
	      }

	    if ((xlate_codes[j].code == GDK_F2) &&
		(xlate_codes[j].modifier & GDK_SHIFT_MASK))
	      {
		static gint deg = 0;
		deg = (deg + 1) % 4;
		  
		gdk_fb_set_rotation (deg);
		break;
	      }

	    if ((xlate_codes[j].code == GDK_F5) &&
		(xlate_codes[j].modifier & GDK_SHIFT_MASK))
	      {
	        gint x, y;

		for (y=0; y<(gdk_display->fb_height+PAVEL_MAGIC_Y-1)/PAVEL_MAGIC_Y; y++)
		  for (x=0; x<(gdk_display->fb_width+PAVEL_MAGIC_X-1)/PAVEL_MAGIC_X; x++)
		    {
		      struct GdkFB_mem_cursed *ukazovatko;
		      short cislo_parecku;

		      ukazovatko = &gdk_display->fb_mem_cursed[x+(gdk_display->fb_width+PAVEL_MAGIC_X-1)/PAVEL_MAGIC_X*y];

		      cislo_parecku = (ukazovatko->background.pixel%8)*8+(ukazovatko->foreground.pixel%8);
		      if (ukazovatko->background.pixel==ukazovatko->foreground.pixel)
		        cislo_parecku = 8*COLOR_BLACK+COLOR_WHITE;

		      if (ukazovatko->foreground.pixel < 8)
			color_set(cislo_parecku, NULL); 
		      else 
			color_set(cislo_parecku|A_BOLD, NULL); 

		      mvaddch(y, x, ukazovatko->offset_y+(ukazovatko->offset_y<10 ? '0' : 'A'-10));
		    }
		refresh();
		while (getch()==-1);
		gdk_fb_redraw_all ();
		break;
	      }

	    if ((xlate_codes[j].code == GDK_F6) &&
		(xlate_codes[j].modifier & GDK_SHIFT_MASK))
	      {
	        gtk_hierarchy_dump();
		break;
	      }

	    if ((xlate_codes[j].code == GDK_Tab) &&
		(modifier & GDK_MOD1_MASK))
	      {
	        GdkWindowObject *private = GDK_WINDOW_P (_gdk_parent_root);
		GList *g_list;

		g_list = g_list_find (private->children, gdk_fb_window_find_focus());
		if (!g_list)
		  g_list = private->children;
		if (g_list)
		  g_list = g_list->next;
		while (g_list && GDK_WINDOW_TOPLEVEL
		       != ((GdkWindowFBData *)GDK_WINDOW_P(g_list->data)->impl)->drawable_data.window_type)
		  g_list = g_list->next;
		if (!g_list)
		  g_list = private->children;
		while (g_list && GDK_WINDOW_TOPLEVEL
		       != ((GdkWindowFBData *)GDK_WINDOW_P(g_list->data)->impl)->drawable_data.window_type)
		  g_list = g_list->next;

		if (g_list)
		  {
		    gdk_fb_window_send_crossing_events (NULL, g_list->data, GDK_CROSSING_NORMAL);
		    gdk_window_raise (g_list->data);
		  }
		break;
	      }

	    if ((xlate_codes[j].code == GDK_F4) &&
		(xlate_codes[j].modifier & GDK_SHIFT_MASK))
	      {
	        GdkWindow *gdkwindow;

		gdkwindow = gdk_fb_window_find_focus();
		if (gdkwindow)
		  {
		    GdkEvent *event = gdk_event_new (GDK_DELETE);
		    void gtk_main_do_event (GdkEvent *event);

		    /* We need to destroy the child - who knows why? */
		    if (GDK_WINDOW_P (gdkwindow)->children)
		      gdkwindow = GDK_WINDOW (GDK_WINDOW_P (gdkwindow)->children->data);

		    event->any.type = GDK_DELETE;
		    event->any.window = g_object_ref (gdkwindow);
		    event->any.send_event = TRUE;

		    gtk_main_do_event (event);
		    gdk_event_free (event);
		  }
		/* It may crash if we continue here. */
		return TRUE;
	      }

	    if ((xlate_codes[j].code == GDK_F8) &&
		(xlate_codes[j].modifier & GDK_SHIFT_MASK))
	      exit (1);

	    
	    gdk_fb_handle_key (xlate_codes[j].code,
			       xlate_codes[j].code,
			       xlate_codes[j].modifier,
			       0,
			       NULL,
			       0,
			       FALSE);
	    gdk_fb_handle_key (xlate_codes[j].code,
			       xlate_codes[j].code,
			       xlate_codes[j].modifier,
			       0,
			       NULL,
			       0,
			       TRUE);
	    handled = TRUE;
	    break;
	  }
	}
      if (!handled)
	{
	  char dummy[2];
	  gint len;
	  
	  keycode = xlate_chars[key_pressed].code;
	  if (keycode == 0)
	    keycode = key_pressed;
	  modifier |= xlate_chars[key_pressed].modifier;
	  
	  dummy[0] = keycode;
	  dummy[1] = 0;

	  len = ((keycode < 255) && isprint (keycode)) ? 1 : 0;
	  gdk_fb_handle_key (keycode,
			     keycode,
			     modifier,
			     0,
			     (len)?g_strdup(dummy) : NULL,
			     len,
			     FALSE);
	  gdk_fb_handle_key (keycode,
			     keycode,
			     modifier,
			     0,
			     (len)?g_strdup(dummy) : NULL,
			     len,
			     TRUE);
	}
    }
  return TRUE;
}


static gboolean
xlate_open (GdkFBKeyboard *kb)
{
  cbreak();
  noecho();
  nodelay(stdscr, TRUE);
  keypad(stdscr, TRUE);
 
  kb->fd = 0; 

  /* Turn non-blocking mode otherwise we would get stuck during reading of
   * xterm mouse events by ncurses.
   * We set it here in gdkkeyboard-fb.c as we setup the g_io watch here
   * and it can be handy even during the keyboard events dispatching.
   */
  fcntl(kb->fd, F_SETFL, O_NONBLOCK);

  kb->io = g_io_channel_unix_new (kb->fd);
  kb->io_tag = g_io_add_watch (kb->io,
			       G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
			       xlate_io,
			       kb);
  
  return TRUE;
}

static void
xlate_close (GdkFBKeyboard *kb)
{
  nocbreak();
  echo();
  nodelay(stdscr, FALSE);
  keypad(stdscr, FALSE);
   
  g_source_remove (kb->io_tag);
  g_io_channel_unref (kb->io);
  kb->fd = -1;
  /* don't close kb->fd, it is the tty from gdk_display */
}

static guint
xlate_lookup (GdkFBKeyboard       *kb,
	      const GdkKeymapKey  *key)
{
  if (key->group != 0)
    return 0;
  if (key->level != 0)
    return 0;
  return key->keycode;
}

static gboolean
xlate_translate (GdkFBKeyboard       *kb,
		 guint                hardware_keycode,
		 GdkModifierType      state,
		 gint                 group,
		 guint               *keyval,
		 gint                *effective_group,
		 gint                *level,
		 GdkModifierType     *consumed_modifiers)
{
  if (keyval)
    *keyval = hardware_keycode;
  if (effective_group)
    *effective_group = 0;
  if (level)
    *level = 0;
  if (consumed_modifiers)
    *consumed_modifiers = 0;
  return TRUE;
}

static gboolean
xlate_get_for_keyval (GdkFBKeyboard       *kb,
		      guint                keyval,
		      GdkKeymapKey       **keys,
		      gint                *n_keys)
{
  *n_keys = 1;
  *keys = g_new (GdkKeymapKey, 1);
  (*keys)->keycode = keyval;
  (*keys)->group = 0;
  (*keys)->level = 0;
  return TRUE;
}

static gboolean
xlate_get_for_keycode (GdkFBKeyboard       *kb,
		       guint                hardware_keycode,
		       GdkKeymapKey       **keys,
		       guint              **keyvals,
		       gint                *n_entries)
{
  if (keys)
    {
      *keys = g_new (GdkKeymapKey, 1);
      (*keys)->keycode = hardware_keycode;
      (*keys)->level = 0;
      (*keys)->group = 0;
    }
  if (keyvals)
    {
      *keyvals = g_new (guint, 1);
      **keyvals = hardware_keycode;
    }
  if (n_entries)
    *n_entries = 1;

  return TRUE;
}

