/* GTK - The GIMP Toolkit
 * Copyright (C) 2001 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* 
 * Authors: Alexander Larsson <alexl@redhat.com>
 */

#undef GDK_DISABLE_DEPRECATED
#undef GTK_DISABLE_DEPRECATED

#include "gtkprivate.h"
#include "gtkwindow.h"
#include "gtkmain.h"
#include "gtkintl.h"
#include "gtkobject.h"
#include "gdkdrawable.h"


#ifdef GDK_WINDOWING_FB 
#define DECORATE_WINDOWS
#include "linux-fb/gdkfb.h"
#endif

#ifdef GDK_WINDOWING_CURSED
#define DECORATE_WINDOWS
#include "cursed/gdkfb.h"
#endif

#ifdef DECORATE_WINDOWS

typedef enum
{
  GTK_WINDOW_REGION_TITLE,
  GTK_WINDOW_REGION_MAXIMIZE,
  GTK_WINDOW_REGION_CLOSE,
  GTK_WINDOW_REGION_BR_RESIZE
} GtkWindowRegionType;

typedef struct _GtkWindowRegion GtkWindowRegion;

struct _GtkWindowRegion
{
  GdkRectangle rect;
  GtkWindowRegionType type;
};

typedef enum
{
  RESIZE_TOP_LEFT,
  RESIZE_TOP,
  RESIZE_TOP_RIGHT,
  RESIZE_RIGHT,
  RESIZE_BOTTOM_RIGHT,
  RESIZE_BOTTOM,
  RESIZE_BOTTOM_LEFT,
  RESIZE_LEFT,
  RESIZE_NONE,
} GtkWindowResizeType;


#define GTK_TYPE_WINDOW_DECORATION                  (gtk_window_decoration_get_type ())
#define GTK_WINDOW_DECORATION(obj)                  (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_WINDOW_DECORATION, GtkWindowDecoration))
#define GTK_WINDOW_DECORATION_CLASS(klass)          (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_WINDOW_DECORATION, GtkWindowDecorationClass))
#define GTK_IS_WINDOW_DECORATION(obj)               (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_WINDOW_DECORATION))
#define GTK_IS_WINDOW_DECORATION_CLASS(klass)       (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_WINDOW_DECORATION))
#define GTK_WINDOW_DECORATION_GET_CLASS(obj)        (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_WINDOW_DECORATION, GtkWindowDecorationClass))


typedef struct _GtkWindowDecoration       GtkWindowDecoration;
typedef struct _GtkWindowDecorationClass  GtkWindowDecorationClass;


static GType  gtk_window_decoration_get_type (void) G_GNUC_CONST;


struct _GtkWindowDecoration
{
  GtkWidget widget;

  gint n_regions;
  GtkWindowRegion *regions;

  gint last_x, last_y;
  gint last_w, last_h;
  
  PangoLayout *pango_layout;

  GtkWindowResizeType resize;
  
  guint moving : 1;
  guint closing : 1;
  guint maximizing : 1;
  guint maximized : 1;
  guint maximizable : 1;
  guint decorated : 1;
  guint real_inner_move : 1;
  guint focused : 1;
};

struct _GtkWindowDecorationClass
{
  GtkWidgetClass parent_class;
};

static void gtk_window_decoration_class_init (GtkWindowDecorationClass *klass);
static void gtk_window_decoration_init       (GtkWindowDecoration      *window_decoration);

static GtkWidgetClass *parent_class;

static GType
gtk_window_decoration_get_type (void)
{
  static GType window_decoration_type = 0;

  if (!window_decoration_type)
    {
      static const GTypeInfo window_decoration_info =
      {
	sizeof (GtkWindowDecorationClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_window_decoration_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkWindowDecoration),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_window_decoration_init,
	NULL,		/* value_table */
      };

      window_decoration_type =
	g_type_register_static (GTK_TYPE_WIDGET, "GtkWindowDecoration",
				&window_decoration_info, 0);
    }

  return window_decoration_type;
}

static void gtk_window_decoration_finalize(GObject *object);

static void
gtk_window_decoration_class_init (GtkWindowDecorationClass *class)
{
  GObjectClass *gobject_class;
  GtkWidgetClass *widget_class;

  parent_class = gtk_type_class (gtk_widget_get_type ());
  gobject_class = G_OBJECT_CLASS (class);
  widget_class = GTK_WIDGET_CLASS (class);

  gobject_class->finalize = gtk_window_decoration_finalize;

  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("border_top",
							     _("Top Window Border Width"),
							     _("Width of the top window decoration border"),
							     0,
							     G_MAXINT,
							     15,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("border_left",
							     _("Left Window Border Width"),
							     _("Width of the left window decoration border"),
							     0,
							     G_MAXINT,
							     3,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("border_right",
							     _("Right Window Border Width"),
							     _("Width of the right window decoration border"),
							     0,
							     G_MAXINT,
							     3,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("border_bottom",
							     _("Bottom Window Border Width"),
							     _("Width of the bottom window decoration border"),
							     0,
							     G_MAXINT,
							     3,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("button_size_x",
							     _("Window Decoration Button Width"),
							     _("Width of the window decoration button"),
							     0,
							     G_MAXINT,
							     9,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("button_size_y",
							     _("Window Decoration Button Height"),
							     _("Height of the window decoration button"),
							     0,
							     G_MAXINT,
							     9,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("button_offset_y",
							     _("Window Decoration Button Vertical Offset"),
							     _("Position of the window decoration button in the y direction"),
							     0,
							     G_MAXINT,
							     2,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("resize_button_size_x",
							     _("Window Decoration Resize Button Width"),
							     _("Width of the window decoration resize button"),
							     0,
							     G_MAXINT,
							     3+10,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("resize_button_size_y",
							     _("Window Decoration Resize Button Height"),
							     _("Height of the window decoration resize button"),
							     0,
							     G_MAXINT,
							     3+10,
							     G_PARAM_READWRITE));
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_string ("title_font",
							        _("Window Decoration Title Font"),
							        _("Font of the window decoration title bar"),
#ifndef GDK_WINDOWING_CURSED
							        "Sans 9",
#else
							        "Sans 1",
#endif
							        G_PARAM_READWRITE));
}

static void
gtk_window_decoration_init (GtkWindowDecoration *window_decoration)
{
  GTK_WIDGET_SET_FLAGS (window_decoration, GTK_NO_WINDOW);

  window_decoration->n_regions = 0;
  window_decoration->regions = NULL;
  window_decoration->pango_layout = NULL;
  window_decoration->resize = RESIZE_NONE;
  window_decoration->moving = FALSE;
  window_decoration->decorated = TRUE;
  window_decoration->closing = FALSE;
  window_decoration->maximizing = FALSE;
  window_decoration->maximized = FALSE;
  window_decoration->maximizable = FALSE;
  window_decoration->real_inner_move = FALSE;
}

static void
gtk_window_decoration_finalize (GObject *object)
{
  GtkWindowDecoration *window_decoration = GTK_WINDOW_DECORATION (object);

  g_free (window_decoration->regions);
  window_decoration->regions = NULL;
  window_decoration->n_regions = 0;

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


static void gtk_decorated_window_recalculate_regions      (GtkWindow      *window);
static GtkWindowRegionType gtk_decorated_window_region_type    (GtkWindow      *window,
						 gint            x,
						 gint            y);
static gint gtk_decorated_window_frame_event    (GtkWindow *window,
						 GdkEvent *event);
static gint gtk_decorated_window_button_press   (GtkWidget      *widget,
						 GdkEventButton *event);
static gint gtk_decorated_window_button_release (GtkWidget      *widget,
						 GdkEventButton *event);
static gint gtk_decorated_window_motion_notify  (GtkWidget      *widget,
						 GdkEventMotion *event);
static gint gtk_decorated_window_window_state   (GtkWidget           *widget,
						 GdkEventWindowState *event);
static void gtk_decorated_window_paint          (GtkWidget      *widget,
						 GdkRectangle   *area);
static gint gtk_decorated_window_focus_change   (GtkWidget         *widget,
						 GdkEventFocus     *event);
static void gtk_decorated_window_realize        (GtkWindow   *window);
static void gtk_decorated_window_unrealize      (GtkWindow   *window);

static void
gtk_decoration_free (GtkWindowDecoration *deco)
{
  g_object_unref (GTK_OBJECT (deco));
}

void
gtk_decorated_window_init (GtkWindow   *window)
{
  GtkWindowDecoration *deco;

  deco = g_object_new (GTK_TYPE_WINDOW_DECORATION, NULL);
  g_object_ref (GTK_OBJECT (deco));
  gtk_object_sink (GTK_OBJECT (deco));
  g_object_set_data_full (G_OBJECT (window), "gtk-window-decoration", deco,
			  (GDestroyNotify) gtk_decoration_free);
  gtk_widget_set_parent(GTK_WIDGET(deco), GTK_WIDGET (window));

  gtk_window_set_has_frame (window, TRUE);

  g_signal_connect (window,
		    "frame_event",
		    G_CALLBACK (gtk_decorated_window_frame_event),
		    window);
  g_signal_connect (window,
		    "focus_in_event",
		    G_CALLBACK (gtk_decorated_window_focus_change),
		    window);
  g_signal_connect (window,
		    "focus_out_event",
		    G_CALLBACK (gtk_decorated_window_focus_change),
		    window);
  g_signal_connect (window,
		    "realize",
		    G_CALLBACK (gtk_decorated_window_realize),
		    window);
  g_signal_connect (window,
		    "unrealize",
		    G_CALLBACK (gtk_decorated_window_unrealize),
		    window);
}

static inline GtkWindowDecoration *
get_decoration (GtkWindow *window)
{
  return (GtkWindowDecoration *)g_object_get_data (G_OBJECT (window), "gtk-window-decoration");
}

void 
gtk_decorated_window_calculate_frame_size (GtkWindow *window)
{
  GdkWMDecoration decorations;
  GtkWindowDecoration *deco = get_decoration (window);
  
  if (gdk_window_get_decorations (GTK_WIDGET (window)->window,
				  &decorations))
    {
      if ((decorations & GDK_DECOR_BORDER) &&
	  (decorations & GDK_DECOR_TITLE))
	{
	  deco->decorated = TRUE;
	  if ((decorations & GDK_DECOR_MAXIMIZE) &&
	      (gtk_window_get_type_hint (window) == GDK_WINDOW_TYPE_HINT_NORMAL))
	    deco->maximizable = TRUE;
	}
      else
	deco->decorated = FALSE;
    }
  else
    {
      deco->decorated = (window->type != GTK_WINDOW_POPUP);
      deco->maximizable = (gtk_window_get_type_hint (window) == GDK_WINDOW_TYPE_HINT_NORMAL);
    }

  if (deco->decorated)
    {
      gint border_left, border_top, border_right, border_bottom;

      gtk_widget_style_get (GTK_WIDGET (deco),
			    "border_left", &border_left,
			    "border_top", &border_top,
			    "border_right", &border_right,
			    "border_bottom", &border_bottom,
			    NULL);
      gtk_window_set_frame_dimensions (window, border_left, border_top, border_right, border_bottom);
    }
  else
    gtk_window_set_frame_dimensions (window, 0, 0, 0, 0);

  gtk_decorated_window_recalculate_regions (window);
}

static gboolean
gtk_decorated_window_inner_change (GdkWindow *win,
				   gint x, gint y,
				   gint width, gint height,
				   gpointer user_data)
{
  GtkWindow *window = (GtkWindow *)user_data;
  GtkWidget *widget = GTK_WIDGET (window);
  GtkWindowDecoration *deco = get_decoration (window);

  if (deco->real_inner_move)
    {
      deco->real_inner_move = FALSE;
      return FALSE;
    }

  deco->real_inner_move = TRUE;
  gdk_window_move_resize (widget->window,
			  window->frame_left, window->frame_top,
			  width, height);

  gdk_window_move_resize (window->frame,
			  x - window->frame_left, y - window->frame_top,
			  width + window->frame_left + window->frame_right,
			  height + window->frame_top + window->frame_bottom);
  return TRUE;
}

static void
gtk_decorated_window_inner_get_pos (GdkWindow *win,
				    gint *x, gint *y,
				    gpointer user_data)
{
  GtkWindow *window = (GtkWindow *)user_data;

  gdk_window_get_position (window->frame, x, y);
  
  *x += window->frame_left;
  *y += window->frame_top;
}

static void
gtk_decorated_window_realize (GtkWindow   *window)
{
  GtkWindowDecoration *deco = get_decoration (window);
  GtkWidget *widget = GTK_WIDGET (window);
  PangoFontDescription *font_desc;
  const gchar *title_font;

  g_return_if_fail (GTK_WIDGET_REALIZED (window));

  /* Realize deco to apply all its style settings from its parent (window). */
  gtk_widget_realize (GTK_WIDGET(deco));

  gtk_widget_style_get (GTK_WIDGET (deco),
			"title_font", &title_font,
			NULL);

  deco->pango_layout = gtk_widget_create_pango_layout (widget,NULL);

  font_desc = pango_font_description_from_string(title_font);
  pango_layout_set_font_description (deco->pango_layout, font_desc);
  pango_font_description_free (font_desc);
  
  gdk_fb_window_set_child_handler (window->frame,
				   gtk_decorated_window_inner_change,
				   gtk_decorated_window_inner_get_pos,
				   window);

  /* This is a huge hack to make frames have the same shape as
     the window they wrap */
  gdk_window_shape_combine_mask (window->frame, GDK_FB_USE_CHILD_SHAPE, 0, 0);
}


static void
gtk_decorated_window_unrealize (GtkWindow   *window)
{
  GtkWindowDecoration *deco = get_decoration (window);

  if (deco->pango_layout)
    {
      g_object_unref (deco->pango_layout);
      deco->pango_layout = NULL;
    }
}

static gint
gtk_decorated_window_frame_event (GtkWindow *window, GdkEvent *event)
{
  GtkWindowDecoration *deco = get_decoration (window);
  GtkWidget *widget = GTK_WIDGET (window);
  GdkEventExpose *expose_event;

  switch (event->type)
    {
    case GDK_EXPOSE:
      expose_event = (GdkEventExpose *)event;
      if (deco->decorated)
	gtk_decorated_window_paint (widget, &expose_event->area);
      return TRUE;
      break;
    case GDK_CONFIGURE:
      gtk_decorated_window_recalculate_regions (window);
      break;
    case GDK_MOTION_NOTIFY:
      return gtk_decorated_window_motion_notify (widget, (GdkEventMotion *)event);
      break;
    case GDK_BUTTON_PRESS:
      return gtk_decorated_window_button_press (widget, (GdkEventButton *)event);
      break;
    case GDK_BUTTON_RELEASE:
      return gtk_decorated_window_button_release (widget, (GdkEventButton *)event);
    case GDK_WINDOW_STATE:
      return gtk_decorated_window_window_state (widget, (GdkEventWindowState *)event);
    default:
      break;
    }
  return FALSE;
}

static gint
gtk_decorated_window_focus_change (GtkWidget         *widget,
				   GdkEventFocus     *event)
{
  GtkWindow *window = GTK_WINDOW(widget);
  GtkWindowDecoration *deco = get_decoration (window);
  deco->focused = event->in;
  gdk_window_invalidate_rect (window->frame, NULL, FALSE);
  return FALSE;
}

static gint
gtk_decorated_window_motion_notify (GtkWidget       *widget,
				    GdkEventMotion  *event)
{
  GtkWindow *window;
  GtkWindowDecoration *deco;
  GdkModifierType mask;
  GdkWindow *win;
  gint x, y;
  gint win_x, win_y, win_w, win_h;
  gint border_left, border_top;
  
  window = GTK_WINDOW (widget);
  deco = get_decoration (window);
  
  if (!deco->decorated)
    return TRUE;
  
  win = widget->window;
  gdk_window_get_pointer (window->frame, &x, &y, &mask);

  gtk_widget_style_get (GTK_WIDGET (deco),
			"border_left", &border_left,
			"border_top", &border_top,
			NULL);

  gdk_window_get_position (window->frame, &win_x, &win_y);
  win_x += border_left;
  win_y += border_top;
  
  gdk_window_get_geometry (win, NULL, NULL, &win_w, &win_h, NULL);

  if (deco->moving)
    {
      int dx, dy;
      dx = x - deco->last_x;
      dy = y - deco->last_y;

      _gtk_window_reposition (window, win_x + dx, win_y + dy);
    }

  if (deco->resize != RESIZE_NONE)
    {
      int w, h;
      gint border_right, border_bottom;

      gtk_widget_style_get (GTK_WIDGET (deco),
			    "border_right", &border_right,
			    "border_bottom", &border_bottom,
			    NULL);

      w = win_w;
      h = win_h;
      
      switch(deco->resize) {
      case RESIZE_BOTTOM_RIGHT:
	w = x - (border_left + border_right);
	h = y - (border_top + border_bottom);
	break;
      case RESIZE_RIGHT:
	w = x - (border_left + border_right);
	break;
      case RESIZE_BOTTOM:
	h = y - (border_top + border_bottom);
	break;
      case RESIZE_TOP_LEFT:
      case RESIZE_TOP:
      case RESIZE_TOP_RIGHT:
      case RESIZE_BOTTOM_LEFT:
      case RESIZE_LEFT:
      default:
	g_warning ("Resize mode %d not handled yet.\n", deco->resize);
	break;
      }
      
      if ((w > 0) && (h > 0))
	{
	  _gtk_window_constrain_size (window, w,h, &w, &h);
	  
	  if ((w != win_w) || (h != win_h))
	    gdk_window_resize (widget->window, w, h);
	}
    }

  return TRUE;
}

static GtkWindowRegionType
gtk_decorated_window_region_type (GtkWindow *window, gint x, gint y)
{
  GtkWindowDecoration *deco = get_decoration (window);
  int i;

  for (i=0;i<deco->n_regions;i++)
    {
      if ((x >= deco->regions[i].rect.x) &&
	  (x -  deco->regions[i].rect.x < deco->regions[i].rect.width) &&
	  (y >= deco->regions[i].rect.y) &&
	  (y -  deco->regions[i].rect.y < deco->regions[i].rect.height))
	return deco->regions[i].type;
    }
  return -1;
}

static gint
gtk_decorated_window_button_press (GtkWidget       *widget,
				   GdkEventButton  *event)
{
  GtkWindow *window;
  GtkWindowRegionType type;
  GtkWindowDecoration *deco;
  gint x, y; 

  window = GTK_WINDOW (widget);
  deco = get_decoration (window);

  if (!deco->decorated)
    return TRUE;

  x = event->x;
  y = event->y;

  type = gtk_decorated_window_region_type (window, x, y);

  switch (type)
    {
    case GTK_WINDOW_REGION_TITLE:
      if (!deco->maximized && event->state & GDK_BUTTON1_MASK)
	{
	  deco->last_x = x;
	  deco->last_y = y;
	  deco->moving = TRUE;
	}
      break;
    case GTK_WINDOW_REGION_MAXIMIZE:
      if (event->state & GDK_BUTTON1_MASK)
	deco->maximizing = TRUE;
      break;
    case GTK_WINDOW_REGION_CLOSE:
      if (event->state & GDK_BUTTON1_MASK)
	deco->closing = TRUE;
      break;
    case GTK_WINDOW_REGION_BR_RESIZE:
      if (!deco->maximized)
	{
	  if (event->state & GDK_BUTTON1_MASK)
	    deco->resize = RESIZE_BOTTOM_RIGHT;
	  deco->last_x = x;
	  deco->last_y = y;
	}
      break;
    default:
      break;
    }
  
  return TRUE;
}

#include "cursed/gdkprivate-fb.h"

static gint
gtk_decorated_window_button_release (GtkWidget	    *widget,
				     GdkEventButton *event)
{
  GtkWindow *window;
  GtkWindowRegionType type;
  GtkWindowDecoration *deco;
      
  window = GTK_WINDOW (widget);
  deco = get_decoration (window);

  if (deco->closing)
    {
      type = gtk_decorated_window_region_type (window, event->x, event->y);
      if (type == GTK_WINDOW_REGION_CLOSE)
	{
	  GdkEvent *event = gdk_event_new (GDK_DELETE);

	  event->any.type = GDK_DELETE;
	  event->any.window = g_object_ref (widget->window);
	  event->any.send_event = TRUE;

	  gtk_main_do_event (event);
	  gdk_event_free (event);
	}
    }
  else if (deco->maximizing)
    {
      type = gtk_decorated_window_region_type (window, event->x, event->y);
      if (type == GTK_WINDOW_REGION_MAXIMIZE)
        {
	  if (deco->maximized)
	    gtk_window_unmaximize (window);
	  else
	    gtk_window_maximize (window);
	}
    }
  
  deco->closing = FALSE;
  deco->maximizing = FALSE;
  deco->moving = FALSE;
  deco->resize = RESIZE_NONE;
  return TRUE;
}

static gint
gtk_decorated_window_window_state (GtkWidget	       *widget,
				   GdkEventWindowState *event)
{
  GtkWindow *window;
  GtkWindowDecoration *deco;
  GdkWindowObject *priv;
      
  window = GTK_WINDOW (widget);
  deco = get_decoration (window);
  priv = GDK_WINDOW_OBJECT (window->frame);

  if (event->changed_mask & GDK_WINDOW_STATE_MAXIMIZED)
    {
      if (event->new_window_state & GDK_WINDOW_STATE_MAXIMIZED)
	{
	  int w, h;
	  gint border_left, border_top, border_right, border_bottom;

	  gtk_widget_style_get (GTK_WIDGET (deco),
				"border_left", &border_left,
				"border_top", &border_top,
				"border_right", &border_right,
				"border_bottom", &border_bottom,
				NULL);

	  gdk_window_get_geometry (widget->window, NULL, NULL,
				   &deco->last_w, &deco->last_h, NULL);
	  gdk_window_get_origin (widget->window, &deco->last_x, &deco->last_y);
	  w = gdk_screen_get_width(gdk_screen_get_default()) - (border_left + border_right);
	  h = gdk_screen_get_height(gdk_screen_get_default()) - (border_top + border_bottom);
	  _gtk_window_constrain_size (window, w, h, &w, &h);
	  if (w != deco->last_w || h != deco->last_h)
	    {
	      gint border_left, border_top;

	      gtk_widget_style_get (GTK_WIDGET (deco),
				    "border_left", &border_left,
				    "border_top", &border_top,
				    NULL);
	      _gtk_window_reposition (window, border_left, border_top);
	      gdk_window_resize (widget->window, w, h);
	      deco->maximized = TRUE;
	    }
	}
      else
	{
	  _gtk_window_reposition (window, deco->last_x, deco->last_y);
	  _gtk_window_constrain_size (window, deco->last_w, deco->last_h,
				      &deco->last_w, &deco->last_h);
	  gdk_window_resize (widget->window, deco->last_w, deco->last_h);
	  deco->maximized = FALSE;
	}
    }
  return TRUE;
}


static void gtk_decorated_window_paint_fillstring (GtkWidget   *widget,
						   GdkRectangle *area,
						   GtkStateType border_state,
						   gint         x,
						   gint         y,
						   gint         width,
						   gint         height,
						   const gchar *text,
						   int          length)
{
  gint xc, yc;
  GdkDrawable *drawable = GTK_WINDOW (widget)->frame;
  GtkStyle *style = gtk_widget_get_style (widget);
  GtkWindowDecoration *deco = get_decoration (GTK_WINDOW (widget));

  if (!deco->pango_layout)
    return;

  if (area)
    gdk_gc_set_clip_rectangle (style->fg_gc [border_state], area);

  gtk_paint_flat_box (style, drawable, GTK_STATE_NORMAL,
		      GTK_SHADOW_NONE, area, widget, "base",
		      x, y,
		      width, height);

  pango_layout_set_text (deco->pango_layout, text, length);

  for (xc = x; xc < x+width; xc++)
    for (yc = y; yc < y+height; yc++)
      gdk_draw_layout (drawable,
		       style->fg_gc [border_state],
		       xc, yc,
		       deco->pango_layout);

  if (area)
    gdk_gc_set_clip_rectangle (style->fg_gc [border_state], NULL);
}

static void gtk_decorated_window_paint_fillchar (GtkWidget   *widget,
						 GdkRectangle *area,
						 GtkStateType border_state,
						 gint         x,
						 gint         y,
						 gint         width,
						 gint         height,
						 GdkWChar     character)
{
  GtkWindowDecoration *deco = get_decoration (GTK_WINDOW (widget));
  gchar buf[6];
  gint charlen;

  if (!deco->pango_layout)
    return;

  charlen = g_unichar_to_utf8 (character, buf);

  gtk_decorated_window_paint_fillstring (widget, area, border_state, x, y, width, height, buf, charlen);
}

static void
gtk_decorated_window_paint (GtkWidget    *widget,
			    GdkRectangle *area)
{
  GtkWindow *window = GTK_WINDOW (widget);
  GtkWindowDecoration *deco = get_decoration (window);
  gint x1, y1, x2, y2;
  GtkStateType border_state;

  if (deco->decorated)
    {
      GdkWindow *frame;
      gint width, height;
      gint border_left, border_top, border_right, border_bottom;
      gint button_size_x, button_size_y, button_offset_y;

      frame = window->frame;
      gdk_drawable_get_size (frame, &width, &height);

      gtk_widget_style_get (GTK_WIDGET (deco),
			    "border_left", &border_left,
			    "border_top", &border_top,
			    "border_right", &border_right,
			    "border_bottom", &border_bottom,
			    "button_size_x", &button_size_x,
			    "button_size_y", &button_size_y,
			    "button_offset_y", &button_offset_y,
			    NULL);

      /* Border: */
      if (deco->focused)
	border_state = GTK_STATE_SELECTED;
      else 
	border_state = GTK_STATE_PRELIGHT;

      /* Top */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   border_left, 0,
					   width - border_left - border_right, border_top,
					   0x2500);	/* BOX DRAWINGS LIGHT HORIZONTAL */
      /* Top Left Corner */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   0, 0,
					   border_left, border_top,
					   0x250c);	/* BOX DRAWINGS LIGHT DOWN AND RIGHT */
      /* Left */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   0, border_top,
					   border_left, height - (border_top + border_bottom),
					   0x2502);	/* BOX DRAWINGS LIGHT VERTICAL */
      /* Bottom Left Corner */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   0, height - border_bottom,
					   border_left, border_bottom,
					   0x2514);	/* BOX DRAWINGS LIGHT UP AND RIGHT */
      /* Bottom */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   border_left, height - border_bottom,
					   width - border_left - border_right, border_bottom,
					   0x2500);	/* BOX DRAWINGS LIGHT HORIZONTAL */
      /* Bottom Right Corner */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   width - border_right, height - border_bottom,
					   border_right, border_bottom,
					   0x2518);	/* BOX DRAWINGS LIGHT UP AND LEFT */
      /* Right */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   width - border_right, border_top,
					   border_top, height - (border_top + border_bottom),
					   0x2502);	/* BOX DRAWINGS LIGHT VERTICAL */
      /* Top Right Corner */
      gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					   width - border_right, 0,
					   border_right, border_top,
					   0x2510);	/* BOX DRAWINGS LIGHT DOWN AND LEFT */
      
#ifndef GDK_WINDOWING_CURSED
      gtk_paint_box (widget->style, frame, border_state, 
		     GTK_SHADOW_OUT, area, widget, "base",
		     0, 0, width, height);
      
      gtk_paint_box (widget->style, frame, border_state, 
		     GTK_SHADOW_IN, area, widget, "base",
		     border_left - 2, border_top - 2,
		     width - (border_left + border_right) + 3,
		     height - (border_top + border_bottom) + 3);
#endif

      /* Title */
      if (deco->pango_layout && GTK_WINDOW (widget)->title)
	{
	  gchar *title_delimited;

	  if (area)
	    gdk_gc_set_clip_rectangle (widget->style->fg_gc [border_state], area);

	  title_delimited = g_strdup_printf (" %s ", GTK_WINDOW (widget)->title);
	  pango_layout_set_text (deco->pango_layout, title_delimited, -1);
	  gdk_draw_layout (frame,
			   widget->style->fg_gc [border_state],
			   border_left * 2,
#ifndef GDK_WINDOWING_CURSED
			   1,
#else
			   0,
#endif
			   deco->pango_layout);
	  g_free (title_delimited);
	  if (area)
	    gdk_gc_set_clip_rectangle (widget->style->fg_gc [border_state], NULL);
	}

      if (deco->maximizable)
	{
	  /* Maximize button: */

	  x1 = width - (border_left * 2) - (button_size_x * 2);
	  y1 = button_offset_y;
	  x2 = x1 + button_size_x;
	  y2 = y1 + button_size_y;

	  gtk_decorated_window_paint_fillstring (widget, area, widget->state, 
					         x1, y1, 1, 1, (deco->maximized
						                ? "[\xE2\x86\x93]"	/* 0x2193 DOWN ARROW */
								: "[\xE2\x86\x91]"),	/* 0x2191 UP ARROW */
						 -1);
	}

      if (window->allow_shrink || window->allow_grow)
	{
	  gint resize_button_size_x, resize_button_size_y;

	  gtk_widget_style_get (GTK_WIDGET (deco),
				"resize_button_size_x", &resize_button_size_x,
				"resize_button_size_y", &resize_button_size_y,
				NULL);

	  gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					       width - resize_button_size_x, height - border_bottom,
					       resize_button_size_x, border_bottom,
					       0x2592);	/* FIXME: ??? checker ??? */

	  gtk_decorated_window_paint_fillchar (widget, area, border_state, 
					       width - border_right, height - resize_button_size_y,
					       border_right, resize_button_size_y,
					       0x2592);	/* FIXME: ??? checker ??? */
	}
      
      /* Close button: */
      
      x1 = width - border_left - button_size_x;
      y1 = button_offset_y;

      gtk_decorated_window_paint_fillstring (widget, area, widget->state, 
					     x1, y1, 1, 1, "[X]", -1);
    }
}

static void
gtk_decorated_window_recalculate_regions (GtkWindow *window)
{
  gint n_regions;
  gint width, height;
  GtkWindowRegion *region;
  GtkWindowDecoration *deco = get_decoration (window);
  gint border_left, border_top, border_right, border_bottom;
  gint button_size_x, button_size_y, button_offset_y;

  n_regions = 0;

  if (!deco->decorated)
    return;

  gtk_widget_style_get (GTK_WIDGET (deco),
			"border_left", &border_left,
			"border_top", &border_top,
			"border_right", &border_right,
			"border_bottom", &border_bottom,
			"button_size_x", &button_size_x,
			"button_size_y", &button_size_y,
			"button_offset_y", &button_offset_y,
			NULL);

  n_regions += 2; /* close, Title */
  if (deco->maximizable)
    n_regions += 1;
  if (window->allow_shrink || window->allow_grow)
    n_regions += 2;

  if (deco->n_regions != n_regions)
    {
      g_free (deco->regions);
      deco->regions = g_new (GtkWindowRegion, n_regions);
      deco->n_regions = n_regions;
    }

  width = GTK_WIDGET (window)->allocation.width + (border_left + border_right);
  height = GTK_WIDGET (window)->allocation.height + (border_top + border_bottom);

  region = deco->regions;

  /* Maximize button */
  if (deco->maximizable)
    {
      region->rect.x = width - (border_left * 2) - (button_size_x * 2);
      region->rect.y = button_offset_y;
      region->rect.width = button_size_x;
      region->rect.height = button_size_y;
      region->type = GTK_WINDOW_REGION_MAXIMIZE;
      region++;
    }

  /* Close button */
  region->rect.x = width - border_left - button_size_x;
  region->rect.y = button_offset_y;
  region->rect.width = button_size_x;
  region->rect.height = button_size_y;
  region->type = GTK_WINDOW_REGION_CLOSE;
  region++;
    
  /* title bar */
  region->rect.x = 0;
  region->rect.y = 0;
  region->rect.width = width;
  region->rect.height = border_top;
  region->type = GTK_WINDOW_REGION_TITLE;
  region++;
  
  if (window->allow_shrink || window->allow_grow)
    {
      gint resize_button_size_x, resize_button_size_y;

      gtk_widget_style_get (GTK_WIDGET (deco),
			    "resize_button_size_x", &resize_button_size_x,
			    "resize_button_size_y", &resize_button_size_y,
			    NULL);

      region->rect.x = width - resize_button_size_x;
      region->rect.y = height - border_bottom;
      region->rect.width = resize_button_size_x;
      region->rect.height = border_bottom;
      region->type = GTK_WINDOW_REGION_BR_RESIZE;
      region++;

      region->rect.x = width - border_right;
      region->rect.y = height - resize_button_size_y;
      region->rect.width = border_right ;
      region->rect.height = resize_button_size_y;
      region->type = GTK_WINDOW_REGION_BR_RESIZE;
      region++;
    }
}

void
gtk_decorated_window_move_resize_window (GtkWindow   *window,
					 gint         x,
					 gint         y,
					 gint         width,
					 gint         height)
{
  GtkWidget *widget = GTK_WIDGET (window);
  GtkWindowDecoration *deco = get_decoration (window);
  
  deco->real_inner_move = TRUE;
  gdk_window_move_resize (widget->window,
			  x, y, width, height);
}
#else

void
gtk_decorated_window_init (GtkWindow  *window)
{
}

void 
gtk_decorated_window_calculate_frame_size (GtkWindow *window)
{
}

void
gtk_decorated_window_move_resize_window (GtkWindow   *window,
					 gint         x,
					 gint         y,
					 gint         width,
					 gint         height)
{
  gdk_window_move_resize (GTK_WIDGET (window)->window,
			  x, y, width, height);
}
#endif



